ThonkWindow {
	var <server, <midiOn, <midiInArray, win, controllersArray, goArray, rout, audioSelectBar, recordButtons, playGroup, recordGroup, midiControlDict, loopOn;
	
	*new { arg server, midiOn, midiInArray;
		^super.newCopyArgs(server, midiOn, midiInArray).init;
	}

	*initClass {
		{
			SynthDef("recordAudio2", {arg bufnum;
				DiskOut.ar(bufnum, In.ar(0,2));
			}).writeDefFile;

			SynthDef("waveset", { arg out = 0, bufnum = 0, start = 0, end, rate = 1, dur = 1, amp=0.2, pan = 0, gate = 1; 
				var phasor;
				phasor = Phasor.ar(0, BufRateScale.kr(bufnum) * rate, 0, end - start) + start;
				OffsetOut.ar(out,
					Pan2.ar(BufRd.ar(1, bufnum, phasor)
				 	* EnvGen.ar(Env([0,amp, 0], [dur/2, dur/2], 'sine'), gate, doneAction: 2), pan)
				 );
			}).writeDefFile;
		}.defer(1);
	}

	init {
		server.waitForBoot({
			playGroup = Group.tail(server);
			recordGroup = Group.tail(server);
			
			win = SCWindow("thOnk", Rect(0, 0, 1245, 490));
			win.front;
			
			goArray = Array.newClear(0);
			
			goArray = goArray.add(SCButton(win, Rect(5, 60, 100, 36)));
			goArray.at(0).states = [["Go", Color.black, Color.blue]];
			goArray.at(0).action = {
				this.playLoop;
			};
			
			goArray = goArray.add(SCButton(win, Rect(105, 60, 100, 36)));
			goArray.at(1).states = [["Stop", Color.black, Color.red]];
			goArray.at(1).action = {
				rout.stop;
			};
			
			goArray = goArray.add(SCButton(win,Rect(205, 60, 100, 36)));
			goArray.at(2).states = [["Loop Off",Color.white,Color.black], ["Loop On",Color.black,Color.green]];
			goArray.at(2).action = {arg butt;
				if(butt.value == 1,
				{loopOn = true},{loopOn = false});
			};
			loopOn = false;
			
			goArray = goArray.add(SCStaticText(win, Rect(405, 40, 100, 40)));
			goArray.at(3).font = Font("Arial", 40);
			goArray.at(3).string = "0";
			
			audioSelectBar = AudioSelectBar(win, [5,0], server);
			recordButtons = RecordButtons(win, [600,0], server, recordGroup.nodeID);
			
			controllersArray = Array.newClear(0);
			controllersArray = controllersArray.add(ControllerBlock(win, [5,100], 0.25, 30, 'linear', 0.001, "Playback Length", 1, [0,1,2,3,4,5], midiOn, false));
			controllersArray = controllersArray.add(ControllerBlock(win, [160,100], 0.25, 4, 'exponential', 0.001, "Pitch", 1, [0,1,2,3,4,5], midiOn, false));
			controllersArray = controllersArray.add(ControllerBlock(win, [315,100], 0, 100, 'linear', 0.001, "Time Deviation", 0, [0,2,5], midiOn, false));
			controllersArray = controllersArray.add(ControllerBlock(win, [470,100], 1, 200, 'linear', 1, "WaveSets Per", 1, [0,1,2,3,4,5], midiOn, false));
			controllersArray = controllersArray.add(ControllerBlock(win, [625,100], 1, 50, 'linear', 1, "Repetitions", 1, [0,1,2,3,4,5], midiOn, false));
			controllersArray = controllersArray.add(ControllerBlock(win, [780,100], -9, 50, 'linear', 0.001, "Overlaps", 0, [0,1,2,3,4,5], midiOn, true));
			controllersArray = controllersArray.add(ControllerBlock(win, [935,100], -1, 1, 'linear', 0.001, "Pan", 0, [0,1,2,3,4,5], midiOn, false));
			controllersArray = controllersArray.add(ControllerBlock(win, [1090,100], 0, 1, 'linear', 0.001, "Volume", 0.5, [0,1,2,3,4,5], midiOn, false));
			
			
			//midi control stuff
			
			midiControlDict = (
				midiInArray.at(0): {arg val; controllersArray.at(0).getMidi1(val)},
				midiInArray.at(1): {arg val; controllersArray.at(0).getMidi2(val)},
				midiInArray.at(2): {arg val; controllersArray.at(1).getMidi1(val)},
				midiInArray.at(3): {arg val; controllersArray.at(1).getMidi2(val)},
				midiInArray.at(4): {arg val; controllersArray.at(2).getMidi1(val)},
				midiInArray.at(5): {arg val; controllersArray.at(2).getMidi2(val)},
				midiInArray.at(6): {arg val; controllersArray.at(3).getMidi1(val)},
				midiInArray.at(7): {arg val; controllersArray.at(3).getMidi2(val)},
				midiInArray.at(8): {arg val; controllersArray.at(4).getMidi1(val)},
				midiInArray.at(9): {arg val; controllersArray.at(4).getMidi2(val)},
				midiInArray.at(10): {arg val; controllersArray.at(5).getMidi1(val)},
				midiInArray.at(11): {arg val; controllersArray.at(5).getMidi2(val)},
				midiInArray.at(12): {arg val; controllersArray.at(6).getMidi1(val)},
				midiInArray.at(13): {arg val; controllersArray.at(6).getMidi2(val)},
				midiInArray.at(14): {arg val; controllersArray.at(7).getMidi1(val)},
				midiInArray.at(15): {arg val; controllersArray.at(7).getMidi2(val)}
			);
			
			if (midiOn == true,
			{	
				MIDIIn.connect; 	
				MIDIIn.control = { arg src, chan, num, val; 
					midiControlDict[num].(val)
				};
			});
		})	
	}
	
	playLoop {
		
		
		rout = Routine({ 
			var window, buttonArray, startTime, currentTime, playBackLength;
			var a, e, rep, dur, rate, pan, amp;
			var startframe, endframe, percent, overlaps;
			var crossings;
			
			//1.wait;
			
			controllersArray.size.do{arg i;
				controllersArray.at(i).gotAGo;
			};
			crossings = audioSelectBar.getCrossings;
			playBackLength = controllersArray.at(0).getBang;
			
			startTime = Main.elapsedTime;
	
			//{controllersArray.at(0).disableAll}.defer;
			loop({
				currentTime = (Main.elapsedTime-startTime);
				if (currentTime >= playBackLength,
				{
					if(loopOn == true, 
					{
						startTime = Main.elapsedTime;
						controllersArray.size.do{arg i;
							controllersArray.at(i).gotAGo;
						};
						crossings = audioSelectBar.getCrossings;
						playBackLength = controllersArray.at(0).getBang;
					},
					{
						if (currentTime >= playBackLength, {nil.alwaysYield});
					});
				},
				{
					{goArray.at(3).string = (playBackLength-currentTime)}.defer;
					percent = currentTime/playBackLength;
					
					a = audioSelectBar.getCrossings2(percent, controllersArray.at(2).getBang(percent));
		 
					rate = controllersArray.at(1).getBang(percent);
					
					e = a + controllersArray.at(3).getBang(percent).round; 
					if(e>(audioSelectBar.waveSets.numXings-1),{e = audioSelectBar.waveSets.numXings-1});
					
					rep = controllersArray.at(4).getBang(percent).round; 
					
					startframe = audioSelectBar.waveSets.fracXings[a];
					endframe = audioSelectBar.waveSets.fracXings[e];
					
					overlaps = controllersArray.at(5).getBang(percent);
					
					dur = (endframe - startframe) * rep / rate / server.sampleRate; 
					pan = controllersArray.at(6).getBang(percent);
					amp = controllersArray.at(7).getBang(percent);
					
					server.sendBundle(0.2, [ "s_new", 
						"waveset", -1, 1, playGroup.nodeID, 
						\bufnum, audioSelectBar.bufnum, 
						\start, startframe, 
						\end, endframe, 
						\rate, rate,
						\dur, dur, 
						\amp, amp,
						\pan, pan
					]);
					if(overlaps>=0,
					{
						(dur * (1/(overlaps+1))).wait;
					},
					{
						(dur*(1+(overlaps.abs))).wait;
					});
				});
			});
		});
		SystemClock.play(rout);
		
	}

	
}

ControllerBlock {
	var <win, <point, <low, <high, <warp, <step, <string1, <initVal, <whichMenuItems, <midiOn, <isOverlaps, controllerBarA, controllerBarB, numArray, inputNumBar, text1, currentMenuItem;
	var popUp, values, useValues, dragSink;
	
	*new { arg win, point, low, high, warp, step, string1, initVal, whichMenuItems, midiOn, isOverlaps;
		^super.newCopyArgs(win, point, low, high, warp, step, string1, initVal, whichMenuItems, midiOn, isOverlaps).init;
	}
	
	init {
		text1 = SCStaticText(win, Rect(point.at(0), point.at(1), 100, 16));
		text1.string = string1.asString;
		
		useValues = Array.newClear(0);
		values = ["single slider", "two sliders", "enter values", "values + 1", "values + 2","user function"];
		whichMenuItems.do{arg i;
			useValues = useValues.add(values.at(i));
		};
		popUp = SCPopUpMenu(win, Rect(point.at(0), point.at(1)+20, 150, 20));
		popUp.items = useValues;
		popUp.action = {arg i;
			currentMenuItem = whichMenuItems.at(i.value);
			switch (currentMenuItem,
				0, 
				{
					inputNumBar.hide;
					controllerBarA.unhide;
					controllerBarB.hide;
					dragSink.hide;
				},
				1,
				{
					inputNumBar.hide;
					controllerBarA.unhide;
					controllerBarB.unhide;
					dragSink.hide;
				},
				2,
				{
					inputNumBar.unhide;
					controllerBarA.hide;
					controllerBarB.hide;
					dragSink.hide;
				},
				3,
				{
					inputNumBar.unhide;
					controllerBarA.unhide;
					controllerBarB.hide;
					dragSink.hide;
				},
				4,
				{
					inputNumBar.unhide;
					controllerBarA.unhide;
					controllerBarB.unhide;
					dragSink.hide;
				},
				5, 
				{
					inputNumBar.hide;
					controllerBarA.hide;
					controllerBarB.hide;
					dragSink.unhide;
				}
			);
		};
		
		inputNumBar = InputNumbersBar(win, [point.at(0), point.at(1)+40]);
		
		if(isOverlaps  == true,
		{
			controllerBarA = ControllerBar2Overlaps(win, [point.at(0), point.at(1)+20], low, high, warp, step, initVal, midiOn);
			controllerBarB = ControllerBar2Overlaps(win, [point.at(0)+75, point.at(1)+20], low, high, warp, step, initVal, midiOn);
		},
		{
			controllerBarA = ControllerBar2(win, [point.at(0), point.at(1)+20], low, high, warp, step, initVal, midiOn);
			controllerBarB = ControllerBar2(win, [point.at(0)+75, point.at(1)+20], low, high, warp, step, initVal, midiOn); });
		dragSink = FunctionSink(win, [point.at(0), point.at(1)+50]);
		
		
		popUp.valueAction = 1;
		popUp.valueAction = 0;
	}
	
	gotAGo {
		if(controllerBarA.isRecording == true, 
		{
			if(controllerBarA.doneRecording == false,
			{
				controllerBarA.initSliderArray;
			},
			{
				controllerBarA.setForLoop;
			})
		});
		if(controllerBarB.isRecording == true, 
		{
			if(controllerBarB.doneRecording == false,
			{
				controllerBarB.initSliderArray;
			},
			{
				controllerBarB.setForLoop;
			})
		});
		this.getNumArray;
	}
	
	getNumArray {
		numArray = inputNumBar.numArray;
	}
	
	disableAll {
		controllerBarA.disableAll;
		controllerBarB.disableAll;
		inputNumBar.disableAll;
	}
	
	enableAll {
		controllerBarA.enableAll;
		controllerBarB.enableAll;
		inputNumBar.enableAll;
	}
	
	getMidi1 {arg val;
		controllerBarA.getMidi(val);
	}
	
	getMidi2 {arg val;
		controllerBarB.getMidi(val);
	}
	
	getBang {arg percent;
		switch (currentMenuItem,
			0, 
			{
				^(controllerBarA.getBang(percent));
			},
			1,
			{
				^(rrand(controllerBarA.getBang(percent), controllerBarB.getBang(percent)));
			},
			2,
			{
				^(numArray.choose.asFloat);
			},
			3,
			{
				^(controllerBarA.getBang(percent)*(numArray.choose.asFloat));
			},
			4,
			{
				^(rrand(controllerBarA.getBang(percent), controllerBarB.getBang(percent))*(numArray.choose.asFloat));
			},
			5,
			{
				^dragSink.getBang(percent);
			}
		);
		
	}
}

FunctionSink {
	var <win, <point, dragSink, text, function;
	
	*new { arg win, point;
		^super.newCopyArgs(win, point).init;
	}
	
	init {
		dragSink = SCDragSink(win, Rect(point.at(0),point.at(1), 100, 50));
		dragSink.action = {arg me; function = me.object};
		text = SCStaticText(win, Rect(point.at(0),point.at(1)+50, 100, 100));
		text.string = "Drop in a function which may or may not receive a number from 0-1 that represents how far through a pass the software has gone."
	}
	
	getBang { arg percent;
		^function.(percent);
	}
	
	hide {
		dragSink.visible_(false);
		text.visible_(false);
	}
	
	unhide {
		dragSink.visible_(true);
		text.visible_(true);
	}
	
}


ControllerBar2 {
	var <win, <point, <low, <high, <warp, <step, <initVal, <midiOn, text1, slider, text2, currentVal, sliderControl, sliderArray, recButton, <isRecording, automationOnButton, automationOn, inputNumBar, numArray, <hidden, midiOnButton, <doneRecording;
	
	*new { arg win, point, low, high, warp, step, initVal, midiOn;
		^super.newCopyArgs(win, point, low, high, warp, step, initVal, midiOn).init;
	}
	
	init {
		currentVal = 0;
		
		slider = SCSlider(win, Rect(point.at(0), point.at(1)+60, 40, 200));
		sliderControl = ControlSpec(low, high, warp, step);
		slider.action = {arg slider;
			currentVal = slider.value;
			text2.string = sliderControl.map(currentVal);
		};
		
		text2 = SCStaticText(win, Rect(point.at(0), point.at(1)+260, 75, 16));
		text2.string = initVal.asString;
		
		recButton = SCButton(win, Rect(point.at(0), point.at(1)+300, 75, 16));
		
		recButton.states = [["NotRecording",Color.white,Color.black],["Recording",Color.black,Color.red]];
		recButton.action = {arg butt;
			if (butt.value == 1,
			{
				isRecording = true; 
				doneRecording = false;
				automationOnButton.valueAction = 0;
			},{isRecording = false});
		};
		recButton.font = Font("Arial", 9);
		
		isRecording = false;
		
		automationOnButton = SCButton(win, Rect(point.at(0), point.at(1)+320, 75, 16));
		automationOnButton.states = [["Automation Off",Color.white,Color.black],["Automation On",Color.black,Color.green]];
		automationOnButton.action = {arg butt;
			if (butt.value == 1,{automationOn = true},{automationOn = false});
		};
		automationOnButton.font = Font("Arial", 9);
		//automationOnButton.enabled = false;
		automationOn = false;
		
		midiOnButton = SCButton(win, Rect(point.at(0), point.at(1)+340, 75, 16));
		midiOnButton.states = [["Midi On",Color.black,Color.green],["Midi Off",Color.white,Color.black]];
		midiOnButton.font = Font("Arial", 9);
		midiOnButton.action = {arg butt;
			if (butt.value == 1,{midiOn = false},{midiOn = true});
		};
		if(midiOn == false, {midiOnButton.valueAction = 1}); 
		
		slider.valueAction = sliderControl.unmap(initVal);
	}
	
	getMidi {arg val;
		if(midiOn == true,
		{
			val = val/127;
			currentVal = val;
			{text2.string = sliderControl.map(currentVal);
			slider.value = val}.defer;
		});
	}
	
	value {
		^currentVal;
	}
	valueReal {
		^(sliderControl.map(currentVal));
	}
	
	initSliderArray {
		sliderArray = Array.newClear(0);
		doneRecording = true;
	}
	
	addToSliderArray {arg index, num;
		sliderArray.put(index, num);
	}
	
	disableAll {
		slider.enabled_(false);
		text2.enabled_(false);
		recButton.enabled_(false);
		automationOnButton.enabled_(false);
	}
	
	enableAll {
		slider.enabled_(true);
		text2.enabled_(true);
		recButton.enabled_(true);
		automationOnButton.enabled_(true);
	}
	
	hide {
		hidden = true;
		slider.visible_(false);
		text2.visible_(false);
		recButton.visible_(false);
		automationOnButton.visible_(false);
		midiOnButton.visible_(false);
	}
	
	unhide {
		hidden = false;
		slider.visible_(true);
		text2.visible_(true);
		recButton.visible_(true);
		automationOnButton.visible_(true);
		midiOnButton.visible_(true);
	}
	
	setForLoop {
		{automationOnButton.value = 1;
		recButton.value = 0;}.defer;
		automationOn = true;
		isRecording = false;
		
	}	
	
	getBang {arg percent;
		var temp;
		
		if (isRecording == true,
		{
			sliderArray = sliderArray.add(currentVal);
		},
		{
			if (automationOn == true,
			{	
				temp = sliderArray.at(((sliderArray.size-1)*percent).floor);
				currentVal = temp;
				{	
					slider.value = temp;
					text2.string = sliderControl.map(currentVal);
				}.defer;
			});
		});
		^(this.valueReal);
	}
	
}

ControllerBar2Overlaps : ControllerBar2 {
	init { 
		super.init;
		
		sliderControl = ControlSpec(low, high, warp, step);
		slider.action = {arg slider;
			var temp;
			
			currentVal = slider.value;
			temp = sliderControl.map(currentVal);
			if(temp<0, {temp = (temp)/3});
			text2.string = temp;
		};
	}
	
	getBang {arg percent;
	var temp2, temp3;
		if (isRecording == true,
		{
			sliderArray = sliderArray.add(currentVal);
		},
		{
			if (automationOn == true,
			{	
				temp2 = sliderArray.at(((sliderArray.size-1)*percent).floor);
				currentVal = temp2;
				{
					slider.value = temp2;
					temp3 = sliderControl.map(currentVal);
					if(temp3<0, {temp3 = (temp3)/3});
					text2.string = temp3;
				}.defer;
			});
		});
		^(this.valueReal);
	}
	
	getMidi {arg val;
		var temp;
		
		if(midiOn == true,
		{
			val = val/127;
			currentVal = val;
			temp = sliderControl.map(currentVal);
			if(temp<0, {temp = (temp)/3});
			{text2.string = temp;
			slider.value = val}.defer;
		});
	}
	
	valueReal {
		var temp;
		
		temp = sliderControl.map(currentVal);
		if(temp<0, {temp = temp/3});
		^temp;
	}
}

InputNumbersBar {
	var <win, <point, label, inputView, text, <numArray;
	
	*new { arg win, point;
		^super.newCopyArgs(win, point).init;
	}
	
	init {
		label = SCStaticText(win,Rect(point.at(0), point.at(1), 150, 16));
		label.string = "Nums split by commas:";
		inputView = SCTextField(win,Rect(point.at(0), point.at(1)+20, 150, 16));
		inputView.action = {arg input;
			var commas;
		
			numArray = Array.newClear(0);
			input.value;
			commas = input.value.findAll(",");
			if (commas == nil, 
			{numArray = numArray.add(input.value.copyRange(0, input.value.size-1))},
			{
				commas = commas.add(input.value.size);
				numArray = numArray.add(input.value.copyRange(0, commas.at(0)-1));
				(commas.size-1).do{arg i;
					numArray = numArray.add(input.value.copyRange(commas.at(i)+1, commas.at(i+1)-1).asFloat);
				};
				numArray.size.do{arg i;
					if (numArray.at(i) == 0, {numArray.put(i, 1)});
				};
			});
			numArray;
		};
		inputView.valueAction = ("1");
	}
	
	hide {
		label.visible_(false);
		inputView.visible_(false);
	}
	
	unhide {
		label.visible_(true);
		inputView.visible_(true);
	}
	
	disableAll {
		label.enabled_(false);
		inputView.enabled_(false);
	}
	
	enableAll {
		label.enabled_(true);
		inputView.enabled_(true);
	}
}


AudioSelectBar {
	var <win, <point, <server, loadFileButton, bufNames, bufPopUp, bufArray, <bufnum, <waveSets, selecterBar, playButton, selectAllButton, lengthText, playBackLengthInput, <playBackLength, loXing, hiXing, crossingSpec, selecterBarText, stopButton, playID;
	
	*new { arg win, point, server;
		^super.newCopyArgs(win, point, server).init;
	}
	
	init {
		
		//waveSets = Wavesets.new;
		
		bufArray = Array.newClear(0);
		bufNames = Array.newClear(0);
		
		loadFileButton = SCButton(win,Rect(point.at(0)+200, point.at(1), 80, 16));
		loadFileButton.states = [["Load File",Color.black,Color.blue]];
		//loadFileButton.value = 0;
		
		bufPopUp = SCPopUpMenu(win, Rect(point.at(0), point.at(1), 190, 16));
		bufPopUp.items = bufNames;
		
		bufPopUp.action = {arg button;
			bufnum = bufArray.at(button.value.asInt).bufnum;
			waveSets = Wavesets.all.at(bufnum.asSymbol);
			crossingSpec = ControlSpec(0, waveSets.numXings, 1, 1);
			
		};
		
		loadFileButton.action = {
			CocoaDialog.getPaths({ arg paths;
				paths.do({ arg path;
					var shortPath, done;
	
					waveSets = Wavesets.from(path, ((bufNames.size)).asSymbol);
					if(waveSets!=nil,
					{
						bufArray = bufArray.add(Buffer.read(server,path));
						shortPath = path.split.pop;
						shortPath.postln;
						bufNames = bufNames.add(shortPath);
						bufPopUp.items = bufNames;
						crossingSpec = ControlSpec(0, waveSets.numXings-1, 1, 1);
						
						loXing = crossingSpec.map(selecterBar.lo);
						hiXing = crossingSpec.map(selecterBar.hi);
						
						Wavesets.all.postln;
						bufPopUp.value = bufNames.size-1;
						bufnum = bufArray.at(bufArray.size-1).bufnum;
					},
					{
						waveSets = Wavesets.all.at(bufnum.asSymbol);
					});
				})
			},{
				"cancelled".postln;
			});
		};
		
		
		selecterBarText = SCStaticText(win, Rect(point.at(0), point.at(1)+20, 100, 16));
		selecterBarText.string = "Select Audio:";
		selecterBar = SCRangeSlider(win, Rect(point.at(0)+100, point.at(1)+20, 100, 16));
		selecterBar.action = {
			loXing = crossingSpec.map(selecterBar.lo);
			hiXing = crossingSpec.map(selecterBar.hi);
		};
		selecterBar.lo = 0;
		selecterBar.hi = 1;
			
		playButton = SCButton(win,Rect(point.at(0)+200, point.at(1)+20, 80, 16));
		playButton.states = [["Play",Color.black,Color.blue]];
		playButton.action = {
			var crossings, startFrame, endFrame, dur;
			
			crossings = this.getCrossings;
			
			startFrame = waveSets.fracXings[crossings.at(0)];
			endFrame = waveSets.fracXings[crossings.at(1)];
			dur = (endFrame-startFrame)/(server.sampleRate);		
			playID = server.nextNodeID;
			
			server.sendBundle(0.2, [ "s_new", 
				"waveset", playID, 0, 0, 
				\bufnum, bufnum, 
				\start, startFrame,
				\end, endFrame, 
				\rate, 1,
				\dur, dur, 
				\amp, 0.5,
				\pan, 0,
				\gate, 1
			]);
		};
		
		stopButton = SCButton(win,Rect(point.at(0)+200, point.at(1)+40, 80, 16));
		stopButton.states = [["Stop",Color.black,Color.red]];
		stopButton.action = {
			"stop".postln;
		
			server.sendMsg("/n_set", playID, \gate, -1);
		};
		
		/*lengthText = SCStaticText(win, Rect(point.at(0), point.at(1)+40, 100, 16));
		lengthText.string = "Playback Length";
		playBackLengthInput = SCNumberBox(win,Rect(point.at(0)+100, point.at(1)+40, 100, 16));
		playBackLengthInput.action = {arg val;
			playBackLength = val.value;
		};
		playBackLengthInput.valueAction = 1;*/
		
	}
	
	getCrossings {
		^[loXing, hiXing]
	}
	
	getCrossings2 {arg percentage, deviation;
		var actualCrossing, devCrossing;
		
		actualCrossing = ((hiXing-loXing)*percentage).floor;
		devCrossing = actualCrossing+((hiXing-loXing)*(rrand(0,deviation/100))*[1,-1].choose);
		if(devCrossing<loXing,
		{devCrossing = loXing+(loXing-devCrossing)},
		{
			if(devCrossing>hiXing,
			{devCrossing = hiXing-(devCrossing-hiXing)})
		});
		^devCrossing
	}
}

RecordButtons {
	var <win, <point, <server, <groupID, <recordArray, fileName, recordID, recordBuf, recordBufNum;
	
	*new { arg win, point, server, groupID, recordArray;
		^super.newCopyArgs(win, point, server, groupID, recordArray).init;
	}
	
	init {
		recordArray = Array.newClear(4);
		
		recordArray.put(0, SCButton(win, Rect(point.at(0), point.at(1), 190, 16))); 
		recordArray.at(0).states = [["Save Audio As:", Color.black, Color.red]];
		recordArray.at(0).action = {
			CocoaDialog.savePanel({ arg path;
				fileName=path;
				recordArray.at(1).string = path;
			},{
				"cancelled".postln;
			});
		};
		recordArray.put(1, SCStaticText(win, Rect(point.at(0),point.at(1)+20,190,56)));
		//recordArray.at(1).action = {arg field; fileName = field.value};
		recordArray.put(2, SCButton(win, Rect(point.at(0),point.at(1)+80,90,16)));
		recordArray.at(2).states = [["Record",Color.black,Color.red]];
		recordArray.at(2).action = {
				// create an output file for this buffer, leave it open
				server.sendMsg("/b_write", recordBufNum, fileName.asString++".aif", "aiff", "int24", 0, 0, 1);
				recordID = server.nextNodeID;
				server.sendMsg("/s_new", "recordAudio2", recordID, 1, groupID, \bufnum, recordBufNum);
		};
		recordArray.put(3, SCButton(win, Rect(point.at(0)+100,point.at(1)+80,90,16)));
		recordArray.at(3).states = [["Stop",Color.black,Color.blue]];
		recordArray.at(3).action = {
			server.sendBundle(nil, ["/n_free", recordID],["/b_close", recordBufNum]);
		};
		
		recordBuf = Buffer.alloc(server, 65536, 2);
		recordBufNum = recordBuf.bufnum;		
	}
}